--[[
Copyright (C) GtX (Andy), 2018

Interface: 1.5.1.0 b1580

Author: GtX	 @  LS-Modcompany
Date: 15.03.2018

Support only at: https://ls-modcompany.com

History:
Version: 1.0.0.0 - Release Edition @ 27.05.2018

Important:
No changes are to be made to this script without permission from GtX @ https://ls-modcompany.com
This includes converting to other game versions before or after FS/LS17

** PLACEABLE ONLY! **
]]

HandPlantSaplings = {}
HandPlantSaplings.dir = g_currentModDirectory

source(HandPlantSaplings.dir.."scripts/events/HPSFillBackpackEvent.lua")
source(HandPlantSaplings.dir.."scripts/events/HPSPlantingEvent.lua")
source(HandPlantSaplings.dir.."scripts/events/HPSSetBackpackEvent.lua")
source(HandPlantSaplings.dir.."scripts/events/HPSSetEquipmentEvent.lua")
source(HandPlantSaplings.dir.."scripts/events/HPSSetStorageEvent.lua")
source(HandPlantSaplings.dir.."scripts/events/HPSSetUserEvent.lua")

local HandPlantSaplings_mt = Class(HandPlantSaplings, Placeable)
InitObjectClass(HandPlantSaplings, "HandPlantSaplings")

function HandPlantSaplings:new(isServer, isClient, customMt)
	local mt = customMt
	if mt == nil then
		mt = HandPlantSaplings_mt
	end

	local self = Placeable:new(isServer, isClient, mt)
	registerObjectClassName(self, "HandPlantSaplings")

	local i = 0
	local modDescXML = loadXMLFile("modDesc", HandPlantSaplings.dir.."modDesc.xml")
	while true do
	   local text = string.format("modDesc.l10n.text(%d)", i)
	   if not hasXMLProperty(modDescXML, text) then
		  break
	   end
	   local name = getXMLString(modDescXML, text.."#name")
	   g_i18n.globalI18N.texts[name] = g_i18n:getText(name)
	   i = i + 1
	end

	return self
end

function HandPlantSaplings:load(xmlFilename, x,y,z, rx,ry,rz, initRandom)
	if not HandPlantSaplings:superClass().load(self, xmlFilename, x,y,z, rx,ry,rz, initRandom) then
		return false
	end

	local xmlFile = loadXMLFile("TempXML", xmlFilename)

	self.shovel = {}
	self.shovel.node = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.handPlantSaplings.shovel#index"))
	self.shovel.node2 = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.handPlantSaplings.shovel#animIndex"))
	self.shovel.parentNode = getParent(self.shovel.node)
	self.shovel.linkPosition = Utils.getVectorNFromString(Utils.getNoNil(getXMLString(xmlFile, "placeable.handPlantSaplings.shovel#linkPosition"), "-0.35 -1 0.6"), 3)
	self.shovel.linkRotation = Utils.getRadiansFromString(Utils.getNoNil(getXMLString(xmlFile, "placeable.handPlantSaplings.shovel#linkRotation"), "90 100 0"), 3)

	self.backpackVisIndex = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.handPlantSaplings.visNodes#backpack"))
	self.saplingsBackpackVis = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.handPlantSaplings.visNodes#saplingsBackpack"))
	self.saplingsStorageVis = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.handPlantSaplings.visNodes#saplingStorage"))

	self.backpackCapacity = Utils.getNoNil(getXMLInt(xmlFile, "placeable.handPlantSaplings#backpackCapacity"), 10)
	self.storageCapacity = Utils.getNoNil(getXMLInt(xmlFile, "placeable.handPlantSaplings#storageCapacity"), 10)


	if self.shovel.node2 ~= nil then
		self.shovelAnimation = Utils.getNoNil(getXMLBool(xmlFile, "placeable.handPlantSaplings.shovelAnimation#active"), false)
	else
		self.shovelAnimation = false
	end

	--local plantingTimeSeconds = Utils.getNoNil(getXMLFloat(xmlFile, "placeable.handPlantSaplings#plantingTimeSeconds"), 4)
	--self.plantingTime = plantingTimeSeconds * 60

	if self.isClient then
		self.playerTriggerIndex = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.handPlantSaplings.playerTrigger#index"))
		self.placeHolderFilename = getXMLString(xmlFile, "placeable.handPlantSaplings.placeHolder#file")
		self.shovelSound = SoundUtil.loadSample(xmlFile, {}, "placeable.handPlantSaplings.shovelSound", nil, self.baseDirectory)
	end

	delete(xmlFile)

	return true
end

function HandPlantSaplings:finalizePlacement()
	HandPlantSaplings:superClass().finalizePlacement(self)

	self.playerCanInteract = false
	self.inOwnTrigger = false
	self.equipmentInUse = false
	self.currentUser = nil
	self.isFieldOwned = nil
	self.canPlantHere = false
	self.placeHolderState = false
	self.resetShovel = true
	self.isPlanting = false
	self.shovelSoundActive = false
	self.placeHolderState = false
	self.nodeLocation = {0,0,0}
	self.lastColourIndex = 0
	self.lastSapling = nil
	self.adminUsersOnly = false
	self.inputWait = 0
	self.placeShown = true
	self.plantingTime = 240
	self.offsetX = 0
	self.offsetY = 0
	self.offsetZ = 0

	self.backpack = {treeType=0, level=0}
	self.storage = {treeType=0, level=0}

	if self.saplingsBackpackVis ~= nil then
		local numberNodes = getNumOfChildren(self.saplingsBackpackVis)
		self.backpackSaplings = {}
		if numberNodes > 0 then
			for b = 1, numberNodes do
				local sapling = getChildAt(self.saplingsBackpackVis, b-1)
				setVisibility(sapling, false)
				table.insert(self.backpackSaplings, sapling)
			end
		else
			print("ERROR: [HandPlantSaplings.lua] No 'sapling' vis nodes found at "..self.saplingsBackpackVis)
		end
	end

	if self.saplingsStorageVis ~= nil then
		local numberNodes = getNumOfChildren(self.saplingsStorageVis)
		self.storageSaplings = {}
		if numberNodes > 0 then
			for s = 1, numberNodes do
				local sapling = getChildAt(self.saplingsStorageVis, s-1)
				setVisibility(sapling, false)
				table.insert(self.storageSaplings, sapling)
			end
		else
			print("ERROR: [HandPlantSaplings.lua] No 'sapling' vis nodes found at "..self.saplingsStorageVis)
		end
	end

	if self.isClient then
		if self.playerTriggerIndex ~= nil then
			self.playerTrigger = self.playerTriggerIndex
			addTrigger(self.playerTrigger, "playerTriggerCallback", self)
		end

		if self.placeHolderFilename ~= nil then
			local i3dNode = Utils.loadSharedI3DFile(self.placeHolderFilename, self.baseDirectory, false, false, false)
			if i3dNode ~= 0 then
				self.placeHolder = getChildAt(i3dNode, 0)
				setVisibility(self.placeHolder, false)
				link(g_currentMission.terrainRootNode, self.placeHolder)
				delete(i3dNode)
			end
		else
			print("ERROR: [HandPlantSaplings.lua] No 'placeHolder' found, this but be added.")
		end
	end
end

function HandPlantSaplings:delete()
	HandPlantSaplings:superClass().delete(self)
	unregisterObjectClassName(self)

	if self.isClient then
		if self.playerTrigger then
			removeTrigger(self.playerTrigger)
		end

		if self.placeHolderFilename ~= nil then
			Utils.releaseSharedI3DFile(self.placeHolderFilename, self.baseDirectory, true)
		end
	end
end

function HandPlantSaplings:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not HandPlantSaplings:superClass().loadFromAttributesAndNodes(self, xmlFile, key, resetVehicles) then
		return false
	end

	local userSetting = getXMLBool(xmlFile, key..".userSetting#adminUsersOnly")
	if userSetting ~= nil then
		self.adminUsersOnly = userSetting
	end

	local backpackLevel = getXMLInt(xmlFile, key..".backpack#fillLevel")
	local backpackTreeType = getXMLInt(xmlFile, key..".backpack#treeType")
	if backpackLevel ~= nil and backpackTreeType ~= nil then
		local treeType = self:getValidTreeType(backpackTreeType, "backpack")
		self:setBackpack(math.min(backpackLevel, self.backpackCapacity), treeType)
	end

	local storageLevel = getXMLInt(xmlFile, key..".storage#fillLevel")
	local storageTreeType = getXMLInt(xmlFile, key..".storage#treeType")
	if storageLevel ~= nil and storageTreeType ~= nil then
		local treeType = self:getValidTreeType(storageTreeType, "storage")
		self:setStorage(math.min(storageLevel, self.storageCapacity), treeType)
	end

	return true
end

function HandPlantSaplings:getSaveAttributesAndNodes(nodeIdent)
	local attributes, nodes = HandPlantSaplings:superClass().getSaveAttributesAndNodes(self, nodeIdent)

	nodes = nodes .. nodeIdent ..'<userSetting adminUsersOnly="'..tostring(self.adminUsersOnly)..'"/>'
	nodes = nodes .. "\n" .. nodeIdent ..'<backpack fillLevel="'..self.backpack.level..'" treeType="'..self.backpack.treeType..'"/>'
	nodes = nodes .. "\n" .. nodeIdent ..'<storage fillLevel="'..self.storage.level..'" treeType="'..self.storage.treeType..'"/>'

	return attributes, nodes
end

function HandPlantSaplings:readStream(streamId, connection)
	HandPlantSaplings:superClass().readStream(self, streamId, connection)

	if connection:getIsServer() then
		local backLevel = streamReadInt8(streamId)
		local backTreeType = streamReadInt8(streamId)
		self:setBackpack(backLevel, backTreeType, true)

		local storeLevel = streamReadInt8(streamId)
		local storeTreeType = streamReadInt8(streamId)
		self:setStorage(storeLevel, storeTreeType, true)

		local isUsed = streamReadBool(streamId)
		if isUsed then
			local user = readNetworkNodeObject(streamId)
			if user ~= nil then
				self:setEquipmentUsed(isUsed, user, true)
			end
		end

		self.adminUsersOnly = streamReadBool(streamId)
	end
end

function HandPlantSaplings:writeStream(streamId, connection)
	HandPlantSaplings:superClass().writeStream(self, streamId, connection)

	if not connection:getIsServer() then
		streamWriteInt8(streamId, self.backpack.level)
		streamWriteInt8(streamId, self.backpack.treeType)

		streamWriteInt8(streamId, self.storage.level)
		streamWriteInt8(streamId, self.storage.treeType)

		streamWriteBool(streamId, self.equipmentInUse)
		if self.equipmentInUse then
			writeNetworkNodeObject(streamId, self.currentUser)
		end

		streamWriteBool(streamId, self.adminUsersOnly)
	end
end

function HandPlantSaplings:update(dt)
	if self.isClient then
		if self.playerCanInteract then
			local backTreeType, backLevel = self:getBackpackLevel()
			local backTypeName = self:getNameFromType(backTreeType)
			local storeTreeType, storeLevel = self:getStorageLevel()
			local storeTypeName = self:getNameFromType(storeTreeType)
			local storageText = string.format("%s %d %s ( %s )", g_i18n:getText("hps_storageStatus"), g_i18n:getFluid(storeLevel), g_i18n:getText("unit_pieces"), storeTypeName)
			g_currentMission:addExtraPrintText(storageText)

			local inUse, assignedUser = self:getEquipmentState()
			if inUse then
				if assignedUser == g_currentMission.player then
					g_currentMission:addHelpButtonText(g_i18n:getText("hps_return"), InputBinding.HPS_EQUIPMENT)
					if InputBinding.hasEvent(InputBinding.HPS_EQUIPMENT) then
						self:setEquipmentUsed(false, nil)
					end

					if self.shovelActive ~= nil then
						local storageSpace = self:getStorageSpace()
						local backpackSpace = self:getBackpackSpace()
						local storageAdd, storageRemove = 0, 0

						if storageSpace > 0 and backLevel > 0 then
							if backLevel < storageSpace then
								storageAdd = backLevel
							else
								storageAdd = storageSpace
							end
							g_currentMission:addHelpButtonText(string.format(g_i18n:getText("HPS_STORAGE_IN"), g_i18n:getFluid(storageAdd), backTypeName), InputBinding.HPS_STORAGE_IN)
						end

						if storeLevel > 0 and backpackSpace > 0 then
							if storeLevel < backpackSpace then
								storageRemove = storeLevel
							else
								storageRemove = backpackSpace
							end

							g_currentMission:addHelpButtonText(string.format(g_i18n:getText("HPS_STORAGE_OUT"), g_i18n:getFluid(storageRemove), storeTypeName), InputBinding.HPS_STORAGE_OUT)
						end

						if g_gui.currentGui == nil then
							if InputBinding.hasEvent(InputBinding.HPS_STORAGE_IN) then
								if backLevel > 0 then
									if storageSpace > 0 then
										if storeTreeType == 0 or storeTreeType == backTreeType then
											self:setBackpack(backLevel - storageAdd, backTreeType)
											self:setStorage(storeLevel + storageAdd, backTreeType)
										else
											g_currentMission:showBlinkingWarning(string.format(g_i18n:getText("hps_storageTypeWarning"), backTypeName))
										end
									else
										g_currentMission:showBlinkingWarning(g_i18n:getText("hps_fullStorageWarning"))
									end
								else
									g_currentMission:showBlinkingWarning(g_i18n:getText("hps_backpackEmpty"))
								end
							end

							if InputBinding.hasEvent(InputBinding.HPS_STORAGE_OUT) then
								if backLevel < self.backpackCapacity then
									if storeLevel > 0 then
										if backTreeType == 0 or backTreeType == storeTreeType then
											self:setBackpack(backLevel + storageRemove, storeTreeType)
											self:setStorage(storeLevel - storageRemove, storeTreeType)
										else
											g_currentMission:showBlinkingWarning(string.format(g_i18n:getText("hps_backpackTypeWarning"), storeTypeName))
										end
									else
										g_currentMission:showBlinkingWarning(g_i18n:getText("hps_emptyStorageWarning"))
									end
								else
									g_currentMission:showBlinkingWarning(g_i18n:getText("hps_backpackFull"))
								end
							end
						end
					end
				else
					g_currentMission:addExtraPrintText(string.format(g_i18n:getText("hps_inUseByOther"), assignedUser.controllerName))
					if InputBinding.hasEvent(InputBinding.HPS_EQUIPMENT) then
						local _, hasShovel = self:getPlayerShovelState()
						local warningText = ""
						if hasShovel then
							warningText = string.format(g_i18n:getText("hps_wrongStorage"), assignedUser.controllerName)
						else
							warningText = string.format(g_i18n:getText("hps_inUseByOther"), assignedUser.controllerName)
						end
						g_currentMission:showBlinkingWarning(warningText, 5000)
					end
				end
			else
				local backpackText = string.format("%s %d %s ( %s )", g_i18n:getText("hps_equipmentStatus"), g_i18n:getFluid(backLevel), g_i18n:getText("unit_pieces"), backTypeName)
				g_currentMission:addExtraPrintText(backpackText)
				g_currentMission:addHelpButtonText(g_i18n:getText("hps_equip"), InputBinding.HPS_EQUIPMENT)

				if InputBinding.hasEvent(InputBinding.HPS_EQUIPMENT) then
					local userName, hasShovel = self:getPlayerShovelState()
					if self:isAllowedUser() then
						if hasShovel then
							local warningText = string.format(g_i18n:getText("hps_inUseByUser"), userName)
							g_currentMission:showBlinkingWarning(warningText, 5000)
						else
							self:setEquipmentUsed(true, g_currentMission.player)
						end
					else
						local text = string.format(g_i18n:getText("hps_mpUsersWarning"), userName)
						g_gui:showInfoDialog({text=text, okText=g_i18n:getText("button_back")})
					end
				end

				if g_currentMission.missionDynamicInfo.isMultiplayer then
					if g_currentMission.isMasterUser or g_currentMission:getIsServer() then
						g_currentMission:addHelpButtonText(g_i18n:getText("HPS_USER_SET"), InputBinding.HPS_USER_SET)
						if InputBinding.hasEvent(InputBinding.HPS_USER_SET) then
							self:setMultiLock()
						end
					end
				end
			end

			if g_currentMission.controlledVehicle ~= nil then
				self.playerCanInteract = false
			end
		end

		if g_currentMission.controlledVehicle == nil then
			if self.currentUser ~= nil and self.currentUser.hasShovel then
				if self.currentUser == g_currentMission.player then
					g_currentMission:addHelpButtonText(g_i18n:getText("HPS_EQUIP_SHOVEL"), InputBinding.HPS_EQUIP_SHOVEL)
					if not self.isPlanting then
						if InputBinding.hasEvent(InputBinding.HPS_EQUIP_SHOVEL) then
							self:setEquipmenState(self.currentUser, false)
						end
					end
				end
			end
		end
	end
end

function HandPlantSaplings:setMultiLock()
	local title = g_i18n:getText("hps_mpUsersControl")
	local text = ""
	if self.adminUsersOnly then
		text = g_i18n:getText("hps_mpUsersAdmin")
	else
		text = g_i18n:getText("hps_mpUsersAll")
	end
	g_gui:showYesNoDialog({text=text, title=title, callback=self.onClickYes, target=self})
end

function HandPlantSaplings:onClickYes(yes)
	if yes then
		local adminOnly = not self.adminUsersOnly
		if g_currentMission:getIsServer() then
			self.adminUsersOnly = adminOnly
			g_server:broadcastEvent(HPSSetUserEvent:new(self, adminOnly))
		else
			g_client:getServerConnection():sendEvent(HPSSetUserEvent:new(self, adminOnly))
		end
	end
end

function HandPlantSaplings:isAllowedUser()
	if not g_currentMission.missionDynamicInfo.isMultiplayer then
		return true
	end

	if self.adminUsersOnly then
		if g_currentMission.isMasterUser or g_currentMission:getIsServer() then
			return true
		end
	else
		return true
	end

	return false
end

function HandPlantSaplings:setEquipmentUsed(isUsed, user, noEventSend)
	HPSSetEquipmentEvent.sendEvent(self, isUsed, user, noEventSend)

	self.equipmentInUse = isUsed

	if isUsed then
		self.currentUser = user
		local shovel = {}
		shovel.node = self.shovel.node
		link(user.toolsRootNode, self.shovel.node)
		setVisibility(self.shovel.node, false)
		self:shovelMovement(false, false)
		shovel.update = HandPlantSaplings.updateShovel
		shovel.updateTick = HandPlantSaplings.updateTickShovel
		shovel.delete = HandPlantSaplings.deleteShovel
		shovel.draw = HandPlantSaplings.drawShovel
		shovel.onActivate = HandPlantSaplings.activateShovel
		shovel.onDeactivate = HandPlantSaplings.deactivateShovel
		shovel.owner = self
		shovel.static = false
		self.shovelTool = shovel
		self:setEquipmenState(user, false)
		user.hasShovel = true
	else
		if self.currentUser ~= nil then
			self:setEquipmenState(user, true)
			self.currentUser.hasShovel = false
		end
		if self.isClient then
			self:plantingSoundState(false)
			link(self.shovel.parentNode, self.shovel.node)
			setVisibility(self.shovel.node, true)
			self:shovelMovement(true, false)
			if not self.placeShown then
				self.placeShown = true
			end
		end
		self.currentUser = nil
	end

	self:updateVisNodes(not isUsed)
end

function HandPlantSaplings:setEquipmenState(user, force)
	if user ~= nil then
		if (user.currentTool == self.shovelTool) or force then
			user:setToolById(0, true)
		else
			user:setTool(self.shovelTool)
		end
	end
end

function HandPlantSaplings.activateShovel(shovel)
	setVisibility(shovel.node, g_currentMission.player == shovel.owner.currentUser)
	shovel.owner.shovelActive = true
end

function HandPlantSaplings.deactivateShovel(shovel)
	setVisibility(shovel.node, false)
	shovel.owner:setPlaceHolder(false)
	shovel.owner.shovelActive = nil
end

function HandPlantSaplings.deleteShovel(shovel)
	shovel.owner:setPlaceHolder(false)
	shovel.owner:setEquipmentUsed(false, nil, false)
end

function HandPlantSaplings.drawShovel(shovel)
	local currentUser = shovel.owner.currentUser == g_currentMission.player

	if currentUser then
		if not shovel.owner.takeSaplings then
			if not shovel.owner.inOwnTrigger then
				if shovel.owner.isPlanting then
					g_currentMission:addHelpButtonText(g_i18n:getText("hps_plantingSapling"), InputBinding.ACTIVATE_HANDTOOL)
				else
					g_currentMission:addHelpButtonText(g_i18n:getText("hps_plantSapling"), InputBinding.ACTIVATE_HANDTOOL)
				end

				if shovel.owner.placeShown then
					g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_turnOffOBJECT"), g_i18n:getText("HPS_MARKER")), InputBinding.HPS_MARKER)
				else
					g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_turnOnOBJECT"), g_i18n:getText("HPS_MARKER")), InputBinding.HPS_MARKER)
				end
			end
		end
	end
end

function HandPlantSaplings.updateShovel(shovel, dt, allowInput)
	local currentUser = shovel.owner.currentUser == g_currentMission.player

	if allowInput and currentUser then
		if InputBinding.hasEvent(InputBinding.HPS_MARKER) then
			shovel.owner.placeShown = not shovel.owner.placeShown
			shovel.owner:setPlaceHolder(shovel.owner.placeHolderState, true)
		end

		local saplingPallet = shovel.owner:getSaplingPalletInRange(g_currentMission.player.rootNode)
		if saplingPallet ~= nil then
			shovel.owner.takeSaplings = true
		else
			shovel.owner.takeSaplings = false
		end

		if shovel.owner.takeSaplings then
			local backpackSpace = shovel.owner:getBackpackSpace()
			if backpackSpace > 0 then
				local spaceText = string.format(g_i18n:getText("hps_takeSaplings"), math.min(backpackSpace, saplingPallet.fillLevel))
				g_currentMission:addHelpButtonText(spaceText, InputBinding.ACTIVATE_HANDTOOL)
			end

			local canFillBackpack, sameType = false, false
			if shovel.owner.backpack.treeType == 0 or shovel.owner.backpack.treeType == saplingPallet.treeType then
				sameType = true
				if shovel.owner:getBackpackSpace() > 0 then
					canFillBackpack = true
				end
			end

			if InputBinding.hasEvent(InputBinding.ACTIVATE_HANDTOOL) then
				shovel.owner.inputWait = 50
				if canFillBackpack then
					local pallet = networkGetObjectId(saplingPallet)
					if shovel.owner.isServer then
						shovel.owner:setSaplingPalletLevel(pallet)
					else
						g_client:getServerConnection():sendEvent(HPSFillBackpackEvent:new(shovel.owner, pallet))
					end
				else
					if sameType then
						g_currentMission:showBlinkingWarning(g_i18n:getText("hps_backpackFull"))
					else
						local typeName = shovel.owner:getNameFromType(saplingPallet.treeType)
						g_currentMission:showBlinkingWarning(string.format(g_i18n:getText("hps_backpackTypeWarning"), typeName))
					end
				end
			end
		end

		if shovel.owner.inputWait > 0 then
			shovel.owner.inputWait = shovel.owner.inputWait - 1
		end

		local x, y, z = localToWorld(g_currentMission.player.cameraNode, 0, 0, -2.5)
		local dx, dy, dz = localDirectionToWorld(g_currentMission.player.cameraNode, 0, -1, 0)
		raycastClosest(x, y, z, dx, dy, dz, "findGroundRaycastCallback", 10, shovel.owner)

		if shovel.owner.backpack.level > 0 then
			local treeType, level = shovel.owner:getBackpackLevel()
			local typeName = shovel.owner:getNameFromType(treeType)
			local backpackText = string.format("%s %d %s ( %s )", g_i18n:getText("hps_backpackText"), g_i18n:getFluid(level), g_i18n:getText("unit_pieces"), typeName)
			g_currentMission:addExtraPrintText(backpackText)
			if shovel.owner:canDoInput() then
				if InputBinding.isPressed(InputBinding.ACTIVATE_HANDTOOL) then
					if not shovel.owner.currentUser.debugFlightMode then
						if shovel.owner.tooClose == nil then
							if shovel.owner.canPlantHere ~= nil then
								local underTreeLimit = TreePlantUtil.canPlantTree()

								if underTreeLimit and shovel.owner.isFieldOwned then
									if not shovel.owner.resetShovel then
										shovel.owner:useShovel(true)
										shovel.owner:shovelMovement(false, true, shovel.owner.nodeLocation)
									end

									shovel.owner.shovelTimer = shovel.owner.shovelTimer - 1

									if shovel.owner.shovelAnimation then
										if shovel.owner.shovelTimer > 230 then
											setRotation(shovel.owner.shovel.node2, 0, 0, 0)
											shovel.owner.offsetZ = shovel.owner.offsetZ + 0.001*dt
										elseif shovel.owner.shovelTimer < 190 and shovel.owner.shovelTimer > 178 then
											shovel.owner.offsetY = shovel.owner.offsetY + 0.001*dt
										elseif shovel.owner.shovelTimer < 178 and shovel.owner.shovelTimer > 160 then
											shovel.owner.offsetX = shovel.owner.offsetX + 0.001*dt
										elseif shovel.owner.shovelTimer < 150 and shovel.owner.shovelTimer > 120 then
											setRotation(shovel.owner.shovel.node2, 0, 0, -20)
											shovel.owner.offsetZ = 0
										elseif shovel.owner.shovelTimer < 120 and shovel.owner.shovelTimer > 110 then
											setRotation(shovel.owner.shovel.node2, 0, 0, 0)
										elseif shovel.owner.shovelTimer < 110 and shovel.owner.shovelTimer > 98 then
											shovel.owner.offsetX = 0
											shovel.owner.offsetY = 0
											shovel.owner.offsetZ = shovel.owner.offsetZ + 0.001*dt
										elseif shovel.owner.shovelTimer < 88 and shovel.owner.shovelTimer > 70 then
											shovel.owner.offsetY = shovel.owner.offsetY + 0.001*dt
										elseif shovel.owner.shovelTimer < 40 and shovel.owner.shovelTimer > 28 then
											shovel.owner.offsetX = shovel.owner.offsetX + 0.001*dt
										elseif shovel.owner.shovelTimer < 28 and shovel.owner.shovelTimer > 0 then
											setRotation(shovel.owner.shovel.node2, 0, 0, -20)
										end
										setTranslation(shovel.owner.shovel.node2, shovel.owner.offsetX, shovel.owner.offsetY, shovel.owner.offsetZ)
									end

									if shovel.owner.shovelTimer == 0 then
										shovel.owner:shovelMovement(false, false)

										local x, y, z = unpack(shovel.owner.nodeLocation)

										shovel.owner.lastSapling = shovel.owner.nodeLocation
										local backpackLevel = shovel.owner.backpack.level - 1
										if g_server then
											shovel.owner:plantSapling( x, y, z, shovel.owner.backpack.treeType, backpackLevel)
										else
											g_client:getServerConnection():sendEvent(HPSPlantingEvent:new(shovel.owner, x, y, z, shovel.owner.backpack.treeType, backpackLevel))
										end
									end
								end

								if not underTreeLimit then
									g_currentMission:showBlinkingWarning(g_i18n:getText("warning_tooManyTrees"))
								end

								if not shovel.owner.isFieldOwned then
									g_currentMission:showBlinkingWarning(g_i18n:getText("warning_youDontOwnThisField"))
								end
							else
								g_currentMission:showBlinkingWarning(string.format(g_i18n:getText("hps_plantWarning"), typeName))
							end
						else
							shovel.owner.isPlanting = false
						end
					else
						shovel.owner.isPlanting = false
						g_currentMission:showBlinkingWarning(g_i18n:getText("hps_flightActive"))
					end
				else
					if shovel.owner.resetShovel then
						shovel.owner:useShovel(false, shovel.owner.plantingTime)
						shovel.owner:shovelMovement(false, false)
					end
				end
			end
		else
			if shovel.owner.resetShovel then
				shovel.owner:useShovel(false, shovel.owner.plantingTime)
				shovel.owner:shovelMovement(false, false)
			end

			if shovel.owner:canDoInput() then
				if InputBinding.hasEvent(InputBinding.ACTIVATE_HANDTOOL) then
					g_currentMission:showBlinkingWarning(g_i18n:getText("hps_emptyBackpackWarning"))
				end
			end
			g_currentMission:addExtraPrintText(g_i18n:getText("hps_emptyBackpack"))
		end

		if shovel.owner.shovelSoundActive ~= shovel.owner.isPlanting then
			shovel.owner.shovelSoundActive = shovel.owner.isPlanting
			shovel.owner:plantingSoundState(shovel.owner.isPlanting)
		end
	end
end

function HandPlantSaplings:shovelMovement(restore, isDigging, translation)
	if restore then
		setTranslation(self.shovel.node, 0, 0, 0)
		setRotation(self.shovel.node, 0, 0, 0)
		setScale(self.shovel.node, 1, 1, 1)
	else
		if isDigging then
			if translation ~= nil and self.shovelAnimation then
				link(g_currentMission.terrainRootNode, self.shovel.node)
				local _, ry, _ = getRotation(g_currentMission.player.cameraNode)
				setTranslation(self.shovel.node, unpack(translation))
				setRotation(self.shovel.node, 0.5, ry+60, 0)
				setScale(self.shovel.node, 1.5, 1.5, 1.5)
			else
				setTranslation(self.shovel.node, 0, -100, 0)
			end
		else
			link(self.currentUser.toolsRootNode, self.shovel.node)
			setTranslation(self.shovel.node, unpack(self.shovel.linkPosition))
			setRotation(self.shovel.node, unpack(self.shovel.linkRotation))
			setScale(self.shovel.node, 1, 1, 1)

			if self.shovelAnimation then
				setTranslation(self.shovel.node2, 0, 0, 0)
				setRotation(self.shovel.node2, 0, 0, 0)
				self.offsetX = 0
				self.offsetY = 0
				self.offsetZ = 0
			end
		end
	end
end

function HandPlantSaplings:useShovel(state, timer)
	self.resetShovel = state
	self.isPlanting = state
	g_currentMission.player.walkingIsLocked = state
	g_currentMission.player.lockedInput = state

	if timer ~= nil then
		self.shovelTimer = timer
	end
end

function HandPlantSaplings:plantingSoundState(isActive)
	if self.shovelSound ~= nil then
		if isActive then
			if self:getIsShovelUserActive() then
				SoundUtil.playSample(self.shovelSound, 0, 0)
			end
		else
			SoundUtil.stopSample(self.shovelSound)
		end
	end
end

function HandPlantSaplings.updateTickShovel(shovel, dt, allowInput)
end

function HandPlantSaplings:findGroundRaycastCallback(hitObjectId, x, y, z, distance, dx, dy, dz)

	self.nodeLocation = {x,y,z}
	self.isFieldOwned = g_currentMission:getIsFieldOwnedAtWorldPos(x,z)

	local canSearch = false
	if self.lastSapling ~= nil then
		local plantingDistance = Utils.vector3Length(x-self.lastSapling[1], y-self.lastSapling[2], z-self.lastSapling[3])
		if plantingDistance > 2 then
			canSearch = true
		end
	else
		canSearch = true
	end

	if not self.currentUser.debugFlightMode then
		if canSearch then
			if hitObjectId == g_currentMission.terrainRootNode and self.backpack.level > 0 then
				if self:canDoInput() then
					self.canPlantHere = true
					if self.isPlanting then
						self:setPlaceHolder(true, false, 3, x, y, z)
					else
						if self.isFieldOwned then
							self:setPlaceHolder(true, false, 2, x, y, z)
						else
							self:setPlaceHolder(true, false, 1, x, y, z)
						end
					end
				else
					self.canPlantHere = nil
					self:setPlaceHolder(false)
				end
			else
				if self:canDoInput() then
					self.canPlantHere = nil
					self:setPlaceHolder(true, false, 1, x, y, z)
				else
					self.canPlantHere = nil
					self:setPlaceHolder(false)
				end
			end
			self.tooClose = nil
		else
			self:setPlaceHolder(true, false, 1, x, y, z)
			self.tooClose = true
		end
	else
		self:setPlaceHolder(false)
	end

	return true
end

function HandPlantSaplings:setPlaceHolder(active, force, colourIndex, x, y, z)
	if self.placeHolder ~= nil then
		if self.placeHolderState ~= active or force then
			setVisibility(self.placeHolder, active and self.placeShown)
			self.placeHolderState = active
		end

		if active and self.placeShown and not force then
			setTranslation(self.placeHolder, x, y, z)

			if colourIndex ~= self.lastColourIndex then
				if colourIndex == 1 then
					setShaderParameter(self.placeHolder, "colorScale", 0.8069, 0.0097, 0.0097, 1, false)
				elseif colourIndex == 2 then
					setShaderParameter(self.placeHolder, "colorScale", 0.395, 0.925, 0.115, 1, false)
				elseif colourIndex == 3 then
					setShaderParameter(self.placeHolder, "colorScale", 0.098, 0.450, 0.960, 1, false)
				end
				self.lastColourIndex = colourIndex
			end
		end
	end
end

function HandPlantSaplings:getSaplingPalletInRange(player)
	local px, py, pz = getWorldTranslation(player)
	local nearestDistance = 3.0
	local nearestSaplingPallet = nil

	for _, item in pairs(g_currentMission.itemsToSave) do
		local object = item.item
		local validPallet = false
		validPallet = object:isa(SaplingPallet)
		if validPallet then
			local vx, vy, vz = getWorldTranslation(object.nodeId)
			local distance = Utils.vector3Length(px-vx, py-vy, pz-vz)
			if distance < nearestDistance then
				nearestSaplingPallet = object
			end
		end
	end

	return nearestSaplingPallet
end

function HandPlantSaplings:setSaplingPalletLevel(palletToSet)
	if self.isServer then
		local saplingPallet = networkGetObject(palletToSet)
		if saplingPallet ~= nil then
			local palletTreeType = saplingPallet.treeType
			local palletFillLevel = saplingPallet.fillLevel
			local levelToFill = self:getBackpackSpace()
			local newPalletLevel = 0

			if palletFillLevel > levelToFill then
				newPalletLevel = palletFillLevel - levelToFill
			else
				levelToFill = palletFillLevel
			end

			saplingPallet:setFillLevel(newPalletLevel)
			local backpackLevel = self.backpack.level + levelToFill
			self:setBackpack(backpackLevel, palletTreeType)
		end
	end
end

function HandPlantSaplings:plantSapling(xPlace, yPlace, zPlace, treeType, backpackLevel)
	if self.isServer then
		local yRot = math.random() * 2 * math.pi
		TreePlantUtil.plantTree(g_currentMission.plantedTrees, treeType, xPlace, yPlace, zPlace, 0, yRot, 0, 0)
		g_currentMission.missionStats:updateStats("plantedTreeCount", 1)

		self:setBackpack(backpackLevel, treeType)
	end
end

function HandPlantSaplings:setBackpack(level, treeType, noEventSend)
	HPSSetBackpackEvent.sendEvent(self, level, treeType, noEventSend)

	if level > 0 then
		self.backpack.treeType = treeType
	else
		self.backpack.treeType = 0
	end

	self.backpack.level = level

	if self.backpackSaplings ~= nil then
		local numNodes = #self.backpackSaplings
		for b=1, numNodes do
			setVisibility(self.backpackSaplings[b], b <= level)
		end
	end
end

function HandPlantSaplings:setStorage(level, treeType, noEventSend)
	HPSSetStorageEvent.sendEvent(self, level, treeType, noEventSend)

	if level > 0 then
		self.storage.treeType = treeType
	else
		self.storage.treeType = 0
	end

	self.storage.level = level

	if self.storageSaplings ~= nil then
		local numNodes = #self.storageSaplings
		for s=1, numNodes do
			setVisibility(self.storageSaplings[s], s <= level)
		end
	end
end

function HandPlantSaplings:updateVisNodes(notUsed)
	if self.isClient then
		if self.backpackVisIndex ~= nil then
			setVisibility(self.backpackVisIndex, notUsed)
		end

		if self.saplingsBackpackVis ~= nil then
			setVisibility(self.saplingsBackpackVis, notUsed)
		end
	end
end

function HandPlantSaplings:canDoInput()
	if not self.takeSaplings and not self.inOwnTrigger and self.inputWait <= 0 then
		return true
	end
	return false
end

function HandPlantSaplings:getIsShovelUserActive()
	if self.equipmentInUse and self.currentUser == g_currentMission.player and not g_gui:getIsGuiVisible() then
		return true
	end
	return false
end

function HandPlantSaplings:getBackpackSpace()
	return self.backpackCapacity - self.backpack.level
end

function HandPlantSaplings:getStorageSpace()
	return self.storageCapacity - self.storage.level
end

function HandPlantSaplings:getNameFromType(treeType)
	local treeName = "N/A"

	if treeType > 0 and treeType <= TreePlantUtil.NUM_TREETYPES then
		treeName = TreePlantUtil.treeTypeIndexToDesc[treeType].nameI18N
	end

	return treeName
end

function HandPlantSaplings:getValidTreeType(treeType, text)
	if treeType > 0 and treeType > TreePlantUtil.NUM_TREETYPES then
		if text ~= nil then
			print(string.format("Warning: [HandPlantSaplings.lua] Unknown treeType '%s' found in %s. Loading default treeType '1' instead.", treeType, text))
		end
		return 1
	end
	return treeType
end

function HandPlantSaplings:getBackpackLevel()
	return self.backpack.treeType, self.backpack.level
end

function HandPlantSaplings:getStorageLevel()
	return self.storage.treeType, self.storage.level
end

function HandPlantSaplings:getEquipmentState()
	local inUse, assignedUser = false, nil

	if self.equipmentInUse then
		if self.currentUser ~= nil then
			inUse = true
			assignedUser = self.currentUser
		end
	end

	return inUse, assignedUser
end

function HandPlantSaplings:getPlayerShovelState()
	if g_currentMission.player == nil then
		return
	end

	local name = g_currentMission.player.controllerName
	local hasShovel = g_currentMission.player.hasShovel

	return name, hasShovel
end

function HandPlantSaplings:playerTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
	if (g_currentMission.controlPlayer and g_currentMission.player and otherId == g_currentMission.player.rootNode) then
		if (onEnter) then
			self.playerCanInteract = true
			if self.equipmentInUse then
				self.inOwnTrigger = g_currentMission.player == self.currentUser
			else
				self.inOwnTrigger = true
			end
		elseif (onLeave) then
			self.playerCanInteract = false
			self.inOwnTrigger = false
		end
	end
end

function HandPlantSaplings:onSell()
	HandPlantSaplings:superClass().onSell(self)

	if self.isServer then
		if self.totalFillTypeSellPrice > 0 then
			g_currentMission:addSharedMoney(self.totalFillTypeSellPrice, "other")
			g_currentMission:addMoneyChange(self.totalFillTypeSellPrice, FSBaseMission.MONEY_TYPE_SINGLE, true, g_i18n:getText("finance_other"))
		end
	end
end

function HandPlantSaplings:canBeSold()
	local warning = ""

	if self.currentUser ~= nil then
		warning = string.format(g_i18n:getText("hps_sellWarningOne"), self.currentUser.controllerName)
		return false, warning
	end

	warning = g_i18n:getText("hps_sellWarningTwo")
	local totalFillLevel = 0
	self.totalFillTypeSellPrice = 0

	if (self.storage.level > 0) or (self.backpack.level > 0) then
		totalFillLevel = self.backpack.level + self.storage.level
		local price = totalFillLevel * 0.80
		warning = string.format("%s (%d %s) - %s: %s\n", warning, g_i18n:getFluid(totalFillLevel), g_i18n:getText("unit_pieces"), g_i18n:getText("ui_sellValue"), g_i18n:formatMoney(price, 0, true, true))
		self.totalFillTypeSellPrice = self.totalFillTypeSellPrice + price
	end

	if totalFillLevel > 0 then
		return true, warning
	end

	return true, nil
end

registerPlaceableType("HandPlantSaplings", HandPlantSaplings)